/* main.c */

#include <SDL.h>
#include <vorbis/vorbisfile.h>
#include <math.h>
#include "main.h"
#include "script.h"
#include "input.h"
#include "attract.h"

#define PICS 14

/* function prototypes */
void write_pnm(Uint8 *);
void audio_callback(void *, Uint8 *, int);
int decode_ogg(OggVorbis_File *, char **);
void render(SDL_Surface *);
void noise(im_t *);
void triangle(im_t *);
void crossfade(im_t *, im_t *, double);
void shutdown(void);
void kyrpele(im_t *);
void nothing(im_t *);

/* data and variables */
extern im_t *pics[SCRIPT_MAX_PICS];
extern int pics_n;
extern input_t tuherrus_m[INPUT_LEN],suodatin_m[INPUT_LEN],displace_m;
int metaframes;
im_t *iabuf,*ibbuf,*nsbuf;
im_t *kyrpe;
im_t *final;
double tuherrus[INPUT_NUM][INPUT_LEN],suodatin[INPUT_NUM][INPUT_LEN],displace[INPUT_NUM][INPUT_LEN];
char *audiobuf;
int audiolen,audiooff;
int renderonly;


int
main(int argc, char *argv[])
{
  SDL_Surface *surface;
  SDL_AudioSpec spec;
  OggVorbis_File vf;
  SDL_Event event;
  Uint32 last, t, flags, startticks;
  char name[50], *script_path;
  FILE *f;
  int i;
  
  /* parse command line arguments */
  flags=SDL_HWSURFACE|SDL_HWPALETTE;
  renderonly=0;

  for (i=1;i<argc;i++)
    {
      if (!strcmp("-fullscreen",argv[i])) flags|=SDL_FULLSCREEN;
      else if (!strcmp("-render",argv[i])) renderonly=1;
      else script_path = argv[i];
    }
  
  /* load pictures */
  for (pics_n=0; pics_n<PICS; pics_n++)
    {
      sprintf(name, "data/pic%d.pgm", pics_n);
      if ((f=fopen(name, "r")) == NULL)
        {
          fprintf(stderr, "error: couldn't open '%s'\n", name);
          return 1;
        }
      pics[pics_n] = im_new_rotozoom(pnm_read(f));
      fclose(f);
      fprintf(stderr, "%s: %dx%d\n", name, pics[pics_n]->w, pics[pics_n]->h);
    }

  f = fopen("data/pic7.pgm","r");
  kyrpe = pnm_read(f);
  fclose(f);

  /* load music */
#if 1
  if (ov_fopen("data/musa.ogg", &vf))
    {
      fprintf(stderr, "error: couldn't open musa.ogg\n");
      return 1;
    }

  if ((audiolen = decode_ogg(&vf,&audiobuf)) == -1)
    {
      fprintf(stderr, "error: unable to read musa.ogg\n");
      return 1;
    }
#endif
  /*...*/
  script_init();
  //script_load(script_path);
  script_load("data/skripti");
  iabuf = im_new(BUF_WIDTH,BUF_HEIGHT,1);
  ibbuf = im_new(BUF_WIDTH,BUF_HEIGHT,1);
  nsbuf = im_new(BUF_WIDTH,BUF_HEIGHT,1);
  final = im_new(SCREEN_WIDTH,SCREEN_HEIGHT,3);

  input_convert(tuherrus_m,tuherrus); input_smooth(50,tuherrus);
  input_convert(suodatin_m,suodatin); input_smooth(50,suodatin);
//  input_convert(displace_m,displace); input_smooth(50,displace);

  /* initialize video and audio */
  if (SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO | SDL_INIT_AUDIO))
    {
      fprintf(stderr, "SDL_Init(): %s\n", SDL_GetError());
      return 1;
    }

  if ((surface=SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 24, flags)) == NULL)
    {
      SDL_Quit();
      return 1;
    }

  spec.freq = 44100;
  spec.format = AUDIO_S16LSB;
  spec.channels = 2;
  spec.samples = 4096;
  spec.callback = audio_callback;
  spec.userdata = NULL;
#if 1
  if (SDL_OpenAudio(&spec,NULL))
    {
      fprintf(stderr, "SDL_OpenAudio(): %s\n", SDL_GetError());
      SDL_Quit(); return 1;
    }

  /* start playing music */
  audiooff = 0;
  SDL_PauseAudio(0);
#endif
  /* and enter main loop */
  metaframes = 0;
  startticks = last = t = SDL_GetTicks();

  script_advance(metaframes);

  while (1)
    {
      /* update metaframe counter (100hz) */
      if (!renderonly)
        {
          t = SDL_GetTicks();
          if (t-last >= 10)
            while (last<t)
              {
                last += 10;
                metaframes++;
                script_advance(1);
                input_record();
              }
        }
      else
        {
          metaframes+=2;
          script_advance(2);
        }
      
      /* draw the next frame (TODO: limit framerate) */
      SDL_LockSurface(surface);
      render(surface);
      if (renderonly) write_pnm(surface->pixels);
      SDL_UnlockSurface(surface);
      if (!renderonly) SDL_Flip(surface);

      /* check for program termination */
      while (SDL_PollEvent(&event))
        switch (event.type)
          {
          case SDL_KEYDOWN: if (event.key.keysym.sym == SDLK_ESCAPE) goto terminate; break;
          case SDL_QUIT: goto terminate;
          }

      if (!renderonly)
        fprintf(stderr, "%d %d %d %f\n", (SDL_GetTicks()-startticks)/1000/60, ((SDL_GetTicks()-startticks)/1000)%60, script_advance(0), input_cur(0));
    }

terminate:
  SDL_Quit();
  input_save(stdout, "brightness");
  fprintf(stderr, "%d\n",script_advance(0));

  return 0;
}

void
write_pnm(Uint8 *pixels)
{
  static int n=0;
  int i, *p;
  FILE *f;
  char path[50];
  
  p = final->data;
  for (i=0;i<SCREEN_WIDTH*SCREEN_HEIGHT;i++)
    {
      *p++ = pixels[i*3+0];
      *p++ = pixels[i*3+1];
      *p++ = pixels[i*3+2];
    }

  sprintf(path, "evening-%.5d.ppm", n++);
  f = fopen(path,"w");
  pnm_write(f,final);
  fclose(f);
}

#define CHUNKLEN 4096

int
decode_ogg(OggVorbis_File *vf, char **buf)
{
  char *p;
  int n,len,section,total;
  
  total = ov_pcm_total(vf,-1)*2*2;
  *buf = p = malloc(total);
  len = 0;

  while ((n = ov_read(vf,p,CHUNKLEN,0,2,1,&section)) && total>=CHUNKLEN)
    {
      total -= n;
      len += n;
      p += n;

#if 0
      {
	vorbis_info *info = ov_info(vf,-1);
	fprintf(stderr, "channels=%d rate=%d\n", info->channels, info->rate);
      }
#endif
    }

  return len;
}

void
audio_callback(void *userdata, Uint8 *buf, int len)
{
  (void)userdata;
  
  if (audiooff+len > audiolen)
    {
      while (len--) *buf++ = 0;
      return;
    }
  
  while (len--)
    *buf++ = audiobuf[audiooff++];
}

void
render(SDL_Surface *surface)
{
  if (metaframes>=0 && metaframes<600)
    {
      noise(iabuf);
      //      kyrpele(iabuf);
      //      triangle(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=600 && metaframes<1050)
    {
      double t = (double)(metaframes-600)/450.0;
      noise(iabuf);
      script_draw(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=1050 && metaframes<3000)
    {
      script_draw(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=3000 && metaframes<3500)
    {
      double t = (double)(metaframes-3000)/500.0;
      triangle(ibbuf);
      script_draw(iabuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=3500 && metaframes<4000)
    {
      double t = (double)(metaframes-3500)/500.0;
      triangle(iabuf);
      script_draw(ibbuf);
      im_crossfade(iabuf,ibbuf,sin(t*M_PI));
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=4000 && metaframes<4300)
    {
      triangle(iabuf);
      noise(ibbuf);
      im_crossfade(iabuf,ibbuf,0.1);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=4300 && metaframes<4450)
    {
      double t = (double)(metaframes-4300)/150.0;
      triangle(iabuf);
      noise(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=4450 && metaframes<7000)
    {
      script_draw(iabuf);
      noise(ibbuf);
      im_crossfade(iabuf,ibbuf,0.2);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=7000 && metaframes<7500)
    {
      double t = (double)(metaframes-7000)/500.0;
      script_draw(ibbuf);
      noise(iabuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=7500 && metaframes<7900)
    {
      script_draw(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=7900 && metaframes<8100)
    {
      double t = (double)(metaframes-7900)/200.0;
      script_draw(iabuf);
      noise(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=8100 && metaframes<8200)
    {
      noise(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=8200 && metaframes<11000)
    {
//      if (input_cur(3)>0)
      if (tuherrus[3][metaframes])
        {
          noise(iabuf);
          kyrpele(ibbuf);
        }
      else
        {
          noise(iabuf);
          script_draw(ibbuf);
        }
      im_crossfade(iabuf,ibbuf,0.8);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=11000 && metaframes<11400)
    {
      double t = (double)(metaframes-11000)/400.0;
      script_draw(iabuf);
      kyrpele(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=11400 && metaframes<11900)
    {
      kyrpele(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=11900 && metaframes<13600)
    {
      kyrpele(iabuf);
      script_draw(ibbuf);
      im_crossfade(iabuf,ibbuf,0.6);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=13600 && metaframes<14100)
    {
      double t = (double)(metaframes-13600)/500.0;
      script_draw(ibbuf);
      kyrpele(iabuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=14100 && metaframes<14400)
    {
      double t = (double)(metaframes-14100)/300.0;
      script_draw(iabuf);
      nothing(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=14400 && metaframes<14500)
    {
      nothing(iabuf);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=14500 && metaframes<14800)
    {
      double t = (double)(metaframes-14500)/300.0;
      script_draw(ibbuf);
      nothing(iabuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else if (metaframes>=14800 && metaframes<15400)
    {
      double t = (double)(metaframes-14800)/600.0;
      script_draw(iabuf);
      nothing(ibbuf);
      im_crossfade(iabuf,ibbuf,t);
      script_render(iabuf,surface->pixels);
    }
  else
    {
      SDL_Quit();
      input_save(stdout, "brightness");
      exit(0);
//      nothing(iabuf);
//      script_render(iabuf,surface->pixels);
    }
}

void
nothing(im_t *buf)
{
  int i;
  
  for (i=0;i<buf->w*buf->h;i++)
    buf->data[i]=0;
}

void
noise(im_t *buf)
{
  double a,b;
  int x,y,i;
  
  for (i=0;i<buf->w*buf->h; i++)
    buf->data[i]=0;
  
  a = (double)metaframes/200.0*M_PI;
  b = (double)metaframes/80.0*M_PI;
  x = 50*cos(sin(a)*M_PI*2)+160;
  y = 60*sin(a)*sin(a)*(cos(a)+1.0)+120;
  im_rotozoom(pics[6],buf, a,50.0+20*sin(b),15.0, x,y, 1.0);
//  buffer_add(buf,nsbuf,1.0);
  
  x = 40*cos(a*1.0)+x;
  y = 40*sin(a*1.0)+y;
  im_rotozoom(pics[rand()%9],buf, -a+M_PI/2,25.0+15*sin(a),125.0, x,y, 1.0);
//  buffer_add(buf,nsbuf,1.0);
  
  buffer_mul(buf,0.00285);
  histogram_equalize(buf,buf);
  buffer_invert(buf);
}

#define TRINUM 8

void
triangle(im_t *buf)
{
  double a,b,c,d,e,f;
  int i,x,y;
  
  for (i=0;i<buf->w*buf->h;i++)
    buf->data[i]=0;
  
  a = (double)metaframes/200.0*M_PI;

  x = 0;
  y = 0;
  im_rotozoom(pics[9],buf,a,1.2,1.2,x,y,1.0);

  f = (double)metaframes/150.0*M_PI;
  b = (double)metaframes/100.0;
  b = fmod(50.0*b,220.0);

  for (i=0; i<TRINUM; i++)
    {
      c = (double)i*2.0*M_PI/(double)TRINUM;
      d = fmod(b+50.0*(double)i,220.0);
      e = d/400.0;
      x = d*cos(c+f);
      y = d*sin(c+f);
      im_rotozoom(pics[10],buf,c+a,e*5.0,0.5,x,y,1.0);
    }
  buffer_clip(buf);
}

/* note: be careful with memory access */
void
kyrpele(im_t *buf)
{
  double a[3],b[3],c[4],*p,*q;
  int i,k,x,y,z,*r,*t,u,v;
 
    {
      a[0] = 0;
      a[1] = 0;
      a[2] = 0;
      p=a,q=b;
  
//      c[0] = 0.2+0.0005*input_cur(0);
      c[0] = 0.2+0.0005*tuherrus[0][metaframes];
//      c[1] = 1.9+0.1*sin(input_cur(0)/640.0*2*M_PI);
      c[1] = 1.9+0.1*sin(tuherrus[0][metaframes]/640.0*2*M_PI);
//      c[2] = 0.73+0.1*sin(input_cur(1)/480.0*2*M_PI);
      c[2] = 0.73+0.1*sin(tuherrus[1][metaframes]/480.0*2*M_PI);
//      c[3] = 0.51+0.0005*input_cur(1);
      c[3] = 0.51+0.0005*tuherrus[1][metaframes];
  
//      if (input_cur(3)>0)
      if (tuherrus[3][metaframes]>0)
        {
          v = metaframes*20;
//          u = (int)input_cur(0)+metaframes;
          u = (int)tuherrus[0][metaframes]+metaframes;
        }
      else
        {
          u = metaframes*20;
//          v = (int)input_cur(0)+metaframes;
          v = (int)tuherrus[0][metaframes]+metaframes;
        }

      memset(buf->data,0,buf->w*buf->h*sizeof(int));

      for (i=0; i<1500; i++)
        {
//      rk(0.01,3,lorenz,c, 0.0,p,q);
          pickover(c,p,q);

          x = (int)((50.0*q[0]/(q[2]==0.0?0.000001:q[2])));
          y = (int)((50.0*q[1]/(q[2]==0.0?0.000001:q[2])));
          z = (int)(fabs(sin(0.1*q[2]))*400);

#if 0
          x%=320;
          y%=200;
          if (x<0) x=320+x;
          if (y<0) y=200+x;
#endif
	  x=abs(x),y=abs(y),z=abs(z);

//      if (x>0 && x<320 && y>0 && y<237)
            {
              r = buf->data + ((y+v)%239)*320+(x+u)%320;
              t = kyrpe->data + ((y+v)%(kyrpe->h-1))*kyrpe->w + (u%kyrpe->w);
              for (k=0; k<(z%120)*2; k++)
                {
                  *r += (*t++)>>1;
                  r++;
                }
            }

          if (p==a) { p=b,q=a; }
          else      { p=a,q=b; }

//          if (input_cur(3)>0)
          if (tuherrus[3][metaframes]>0)
            u++;
        }
    }

  buffer_clip(buf);
}
